@@ -167,7 +167,7 @@ function isString(obj) {
  * \return true if the given object is an RVector object.
  */
 function isVector(obj) {
-    return (typeof(obj)==="object" && obj.toString().startsWith("RVector"));
+    return (typeof(obj)==="object" && isOfType(obj, RVector));
 }
 
 /**
@@ -179,6 +179,24 @@ function isValidVector(obj) {
     return (isVector(obj) && obj.isValid() && isNumber(obj.x) && isNumber(obj.y));
 }
 
+/**
+ * Checks if the given object is a color.
+ *
+ * \return true if the given object is an RColor object.
+ */
+function isColor(obj) {
+    return (typeof(obj)==="object" && isOfType(obj, RColor));
+}
+
+/**
+ * Checks if the given object is a valid color.
+ *
+ * \return true if the given object is a valid RColor object.
+ */
+function isValidColor(obj) {
+    return (isColor(obj) && obj.isValid());
+}
+
 /**
  * Checks if the given object is a valid boolean value.
  *
@@ -625,6 +643,15 @@ function isLeaderEntity(obj) {
     return isOfType(obj, RLeaderEntity) || isOfType(obj, RLeaderEntityPointer);
 }
 
+/**
+ * Checks if the given object is a tolerance entity.
+ *
+ * \return true if the given object is a tolerance entity (RToleranceEntity).
+ */
+function isToleranceEntity(obj) {
+    return isOfType(obj, RToleranceEntity) || isOfType(obj, RToleranceEntityPointer);
+}
+
 /**
  * Checks if the given object is an entity of the given type.
  */
@@ -884,6 +911,8 @@ function entityTypeToString(type, plural) {
         return plural ? qsTr("Images") : qsTr("Image");
     case RS.EntityLeader:
         return plural ? qsTr("Leaders") : qsTr("Leader");
+    case RS.EntityTolerance:
+        return plural ? qsTr("Tolerances") : qsTr("Tolerance");
     case RS.EntityLine:
         return plural ? qsTr("Lines") : qsTr("Line");
     case RS.EntityXLine:
@@ -922,6 +951,7 @@ function translateFilterStrings(filterStrings) {
         filterStrings[i] = filterStrings[i].replace("DXF Drawing", qsTr("DXF Drawing"));
         filterStrings[i] = filterStrings[i].replace("DWG Drawing", qsTr("DWG Drawing"));
         filterStrings[i] = filterStrings[i].replace("PDF File", qsTr("PDF File"));
+        filterStrings[i] = filterStrings[i].replace("PDF/A-1B File", qsTr("PDF/A-1B File"));
     }
     return filterStrings;
 }
@@ -1036,20 +1066,6 @@ function getWidgets(widget, ca, allowDuplicates) {
     return ca;
 }
 
-/**
- * Sleeps for the given milliseconds.
- * \param msec the time to sleep
- */
-function sleep(msec) {
-    var date = new Date();
-    var curDate = null;
-    var c = 0;
-    do {
-        curDate = new Date();
-        ++c;
-    } while (curDate - date < msec);
-}
-
 /**
  * Converts radiant to degrees.
  * \param val the value in radiant
@@ -1844,38 +1860,43 @@ function stringToDirectDistanceEntry(relativeZero, cursorPosition, str) {
  * Creates and returns a new entity based on the given shape.
  */
 function shapeToEntity(document, shape) {
-    if (isPointShape(shape)) {
-        return new RPointEntity(document, new RPointData(shape.getPosition()));
-    }
-    else if (isLineShape(shape)) {
-        return new RLineEntity(document, new RLineData(shape));
-    }
-    else if (isRayShape(shape)) {
-        return new RRayEntity(document, new RRayData(shape));
-    }
-    else if (isXLineShape(shape)) {
-        return new RXLineEntity(document, new RXLineData(shape));
-    }
-    else if (isArcShape(shape)) {
-        return new RArcEntity(document, new RArcData(shape));
-    }
-    else if (isCircleShape(shape)) {
-        return new RCircleEntity(document, new RCircleData(shape));
-    }
-    else if (isEllipseShape(shape)) {
-        return new REllipseEntity(document, new REllipseData(shape));
-    }
-    else if (isPolylineShape(shape)) {
-        return new RPolylineEntity(document, new RPolylineData(shape));
-    }
-    else if (isSplineShape(shape)) {
-        return new RSplineEntity(document, new RSplineData(shape));
-    }
-    else if (isTriangleShape(shape)) {
-        return new RSolidEntity(document, new RSolidData(shape));
+    var s = shape;
+    if (isFunction(s.data)) {
+        s = s.data();
     }
 
-    qWarning("shapeToEntity: unknown shape: ", shape);
+    if (isPointShape(s)) {
+        return new RPointEntity(document, new RPointData(s.getPosition()));
+    }
+    else if (isLineShape(s)) {
+        return new RLineEntity(document, new RLineData(s));
+    }
+    else if (isRayShape(s)) {
+        return new RRayEntity(document, new RRayData(s));
+    }
+    else if (isXLineShape(s)) {
+        return new RXLineEntity(document, new RXLineData(s));
+    }
+    else if (isArcShape(s)) {
+        return new RArcEntity(document, new RArcData(s));
+    }
+    else if (isCircleShape(s)) {
+        return new RCircleEntity(document, new RCircleData(s));
+    }
+    else if (isEllipseShape(s)) {
+        return new REllipseEntity(document, new REllipseData(s));
+    }
+    else if (isPolylineShape(s)) {
+        return new RPolylineEntity(document, new RPolylineData(s));
+    }
+    else if (isSplineShape(s)) {
+        return new RSplineEntity(document, new RSplineData(s));
+    }
+    else if (isTriangleShape(s)) {
+        return new RSolidEntity(document, new RSolidData(s));
+    }
+
+    qWarning("shapeToEntity: unknown shape: ", s);
     return undefined;
 }
 
@@ -2373,14 +2394,55 @@ function neutralPath(path) {
     return path;
 }
 
+/**
+ * \return Mapped icon path for the given icon for the current theme or
+ * mapped icon path for a dark theme (-inverse) or the same path or undefined if no such
+ * icon can be found.
+ */
 function autoIconPath(path) {
-    if (RSettings.hasDarkGuiBackground()) {
-        var pathDark = path.replace(/\.svg$/, "-inverse.svg");
-        if (new QFileInfo(pathDark).exists()) {
-            return pathDark;
+    // set theme specific icon:
+    var themePath = RSettings.getThemePath();
+    var themeIconFile = undefined;
+    if (themePath.length>0) {
+        var fi = new QFileInfo(path);
+        var iconFileName = fi.fileName();
+        themeIconFile = themePath + "/icons/" + iconFileName;
+        if (!new QFileInfo(themeIconFile).exists()) {
+            // no SVG found, look up PNG:
+            var iconBaseName = fi.baseName();
+            themeIconFile = themePath + "/icons/" + iconBaseName + ".png";
+            if (!new QFileInfo(themeIconFile).exists()) {
+                // no PNG found, use default icon:
+                themeIconFile = undefined;
+            }
         }
     }
 
+    if (!isNull(themeIconFile)) {
+        // got icon from theme:
+        return themeIconFile;
+    }
+
+    // no theme icon, try dark mode icon:
+    var darkModeIconFile = undefined;
+    if (RSettings.hasDarkGuiBackground()) {
+        darkModeIconFile = path.replace(/\.svg$/, "-inverse.svg");
+        darkModeIconFile = darkModeIconFile.replace(/\.png$/, "-inverse.png");
+        if (!new QFileInfo(darkModeIconFile).exists()) {
+            darkModeIconFile = undefined;
+        }
+    }
+
+    if (!isNull(darkModeIconFile)) {
+        // got dark mode icon:
+        return darkModeIconFile;
+    }
+
+    if (!new QFileInfo(path).exists()) {
+        // given icon does not exist (for icons based on object names, e.g. in options toolbar):
+        return undefined;
+    }
+
     return path;
 }
 
@@ -2401,17 +2463,22 @@ function applyTheme() {
             var postfix = postfixes[i];
             var fn = path + "stylesheet" + postfix + ".css";
             qDebug("trying to load theme stylesheet: ", fn);
+
             if (new QFileInfo(fn).exists()) {
-                var file = new QFile(fn);
-                var flags = new QIODevice.OpenMode(QIODevice.ReadOnly | QIODevice.Text);
-                if (file.open(flags)) {
-                    var textStream = new QTextStream(file);
-                    var allLines = textStream.readAll();
-                    file.close();
-                    allLines = allLines.replace(/url\(/g, "url(" + path);
-                    qApp.styleSheet = qApp.styleSheet + "\n" + allLines;
-                    found = true;
+                var css = readTextFile(fn);
+                if (css.contains("RequiresPlugin:true")) {
+                    // only load theme if plugin loaded:
+                    var pluginId = theme.toUpperCase() + "STYLE";
+                    if (!RPluginLoader.hasPlugin(pluginId)) {
+                        qWarning("Theme not loaded: ", theme);
+                        qWarning("Theme plugin not found:", pluginId);
+                        return;
+                    }
                 }
+
+                css = css.replace(/url\(/g, "url(" + path);
+                qApp.styleSheet = qApp.styleSheet + "\n" + css;
+                found = true;
             }
         }
 
@@ -2428,6 +2495,7 @@ function readTextFile(fileName) {
     var flags = new QIODevice.OpenMode(QIODevice.ReadOnly | QIODevice.Text);
     if (file.open(flags)) {
         var textStream = new QTextStream(file);
+        textStream.setCodec("UTF-8");
         var contents = textStream.readAll();
         file.close();
         return contents;
@@ -2441,6 +2509,7 @@ function writeTextFile(fileName, str) {
     var flags = new QIODevice.OpenMode(QIODevice.WriteOnly | QIODevice.Text);
     if (file.open(flags)) {
         var textStream = new QTextStream(file);
+        textStream.setCodec("UTF-8");
         textStream.writeString(str);
     }
     file.close();
